% For use with Matlab
% Copyright (C) 2019 -- Jérôme Mendes <jermendes@gmail.com> or <jermendes@isr.uc.pt>, Rui Araújo <rui@isr.uc.pt>
% Jérôme Mendes, Francisco Souza, and Rui Araújo. "Online Evolving Fuzzy Control Design: An Application to a CSTR Plant". In Proc. IEEE 15th International Conference on Industrial Informatics (INDIN 2017), 2017.
% The "SEDFLC toolbox" comes with ABSOLUTELY NO WARRANTY;
% In case of publication of any application of this method, please, cite the work:
% Jérôme Mendes, Francisco Souza and Rui Araújo. 
% Online Evolving Fuzzy Control Design: An Application to a CSTR Plant.
% In Proc. IEEE 15th International Conference on Industrial Informatics (INDIN 2017), pages 1-8, Emden, Germany, July 24 - July 17
% DOI: https://doi.org/10.1109/INDIN.2017.8104774
% http://www.isr.uc.pt/~jermendes/    ;     http://www.isr.uc.pt/~rui/

function  [x, Add, rule] = AddMF_ErrorAVG(x, nearestMF, j, MF_center)
%% Description
% In this function, if the Criterion 2 is met, a new  membership function is added on input variable j 
%% Function inputs
% x - Struct with all fuzzy control parameters
% nearestMF - the nearest membership function from candidate membership function
% j - input variable which has the maximal estimated control error (i.e. where the new MF will be added)
% MF_center - position/center of the new (candidate) membership function

%% Function output
% x - Struct with all fuzzy control parameters
% Add - Flag to indicate if the candidate membership function was added
% (Add=1) or not (Add=0)
% rule - index of the new rule

%% Auxiliary variables
n_rules = size( x(j).Rules,1 ); % number of rules on input variable j
tau = x(j).eta; %  the minimal distance beetween MFs for input variable j
rule = -1;
Add = 0;


%% Case 1 - If the candidate membership function is not the first  or last membership function
if nearestMF ~= 1 && nearestMF < n_rules
    
    b_MostActive = x(j).Rules( nearestMF, 2 ); % MF's center of nearest MF  from candidate MF
    b_previous = x(j).Rules( nearestMF-1, 2 );% MF's center of the previous MF of nearest MF 
    b_next = x(j).Rules( nearestMF+1, 2 );% MF's center of the next MF of nearest MF 

    dist = abs(MF_center - b_MostActive) / 2; % distance beetween the candidate and nearest MFs
    dist_previous = abs(MF_center - b_previous) / 2; % distance beetween the candidate and the previous MF of nearest MFs
    dist_next = abs(MF_center - b_next) / 2;  % distance beetween the candidate and the next MF of nearest MFs
    
    %% Criterion 2
    if dist > tau % If the distance beetween the candidate and nearest MFs is bigger than the  minimal distance admissible
        
        if MF_center < b_MostActive && dist_previous > tau
            
            Add = 1; % The candidate membership function will be added - Criterion 2 is met
            
            new_b =  MF_center; % center of the MF
            new_left =  x(j).Rules( nearestMF - 1, 2 ); % lower limit of the new MF
            new_right =  x(j).Rules( nearestMF, 2 ); % upper limit of the new MF
            
            % Changes on the left MF of the new MF
            x(j).Rules( nearestMF - 1, 3 ) = new_b;
            
            % Changes on the right MF of the new MF
            x(j).Rules( nearestMF , 1 ) = new_b;
            
            % update/sort the Rule Struct
            for i = n_rules:-1:nearestMF
                [x(j).Rules( i+1,:)] = x(j).Rules( i,: );
            end
            
            % Insert the new rule
            x(j).Rules(nearestMF, 1 ) = new_left;
            x(j).Rules(nearestMF, 2 ) = new_b;
            x(j).Rules(nearestMF, 3 ) = new_right;
            x(j).Rules(nearestMF, 4 ) = 0.0;
            
            rule = nearestMF; % index of the new rule
        end 
        if MF_center > b_MostActive && dist_next > tau
            
            Add = 1;% The candidate membership function will be added - Criterion 2 is met
            
            new_b =  MF_center;% center of the MF
            new_left = x(j).Rules( nearestMF , 2 );% lower limit of the new MF
            new_right =  x(j).Rules( nearestMF + 1 , 2 );% upper limit of the new MF
            
            % Changes on the previous MF of the new MF
            x(j).Rules( nearestMF , 3 ) = new_b;
            
            % Changes on next MF of the new MF
            x(j).Rules( nearestMF +1, 1 ) = new_b;
            
            % update/sort the Rule Struct
            for i = n_rules:-1:nearestMF+1
                [x(j).Rules(i+1,:)] = x(j).Rules(i,:);
            end
            
            % Insert the new rule
            x(j).Rules(nearestMF+1, 1 ) = new_left;
            x(j).Rules(nearestMF+1, 2 ) = new_b;
            x(j).Rules(nearestMF+1, 3 ) = new_right;
            x(j).Rules(nearestMF+1, 4 ) = 0.0;
            
            rule = nearestMF+1;% index of the new rule 
        end
        
    end
    %% Case 2 - If the candidate membership function is the first membership function
elseif nearestMF == 1
  
    b_MostActive = x(j).Rules( nearestMF , 2 );% MF's center of nearest MF  from candidate MF
    b_next = x(j).Rules( nearestMF+1 , 2 );% MF's center of the next MF of nearest MF 
    
    dist_1 = abs(MF_center - b_MostActive) / 2;% distance beetween the candidate and nearest MFs
    dist_2 = abs(MF_center - b_next) / 2;% distance beetween the candidate and the next MF of nearest MFs
    
    if dist_1 > tau && dist_2 > tau && MF_center > b_MostActive % If the distance beetween the candidate and nearest MFs is bigger than the  minimal distance admissible

        Add = 1;% The candidate membership function will be added - Criterion 2 is met
        
        new_b =  MF_center;% center of the MF
        new_left = x(j).Rules( nearestMF , 2 );% lower limit of the new MF
        new_right = x(j).Rules( nearestMF + 1, 2 ); % upper limit of the new MF
        
        % Changes  on the left MF of the new MF
        x(j).Rules( nearestMF , 3 ) = new_b;

        % Changes on the right MF of the new MF
        x(j).Rules( nearestMF + 1, 1 ) = new_b;
        
        % update/sort the Rule Struct
        for i = n_rules:-1:nearestMF+1
            [x(j).Rules(i+1,:)] = x(j).Rules(i,:);
        end
          
        % Insert the new rule
        x(j).Rules(nearestMF+1, 1 ) = new_left;
        x(j).Rules(nearestMF+1, 2 ) = new_b;
        x(j).Rules(nearestMF+1, 3 ) = new_right;
        x(j).Rules(nearestMF+1, 4 ) = 0.0;
        
        rule = nearestMF+1;% index of the new rule         
    end
    %% Case 3 - If the candidate membership function is the last membership function
else
   
    b_MostActive = x(j).Rules( nearestMF , 2 );% MF's center of nearest MF  from candidate MF
    b_previous= x(j).Rules( nearestMF-1, 2 );% MF's center of the previous MF of nearest MF 
    
    dist_1 = abs(MF_center - b_MostActive) / 2;% distance beetween the candidate and nearest MFs
    dist_2 = abs(MF_center - b_previous) / 2;% distance beetween the candidate and the previous MF of nearest MFs
    
    if dist_1 > tau && dist_2 > tau   && MF_center < b_MostActive % If the distance beetween the candidate and nearest MFs is bigger than the  minimal distance admissible
        
        Add = 1;% The candidate membership function will be added - Criterion 2 is met
        
        new_b =  MF_center;% center of the MF
        
        % Changes  on the left MF of the new MF
        x(j).Rules( nearestMF - 1 , 3 ) = new_b;
        
        % Creat the new MF
        new_left = x(j).Rules( nearestMF - 1 , 2 );% lower limit of the new MF
        new_right = x(j).Rules( nearestMF , 2 );% upper limit of the new MF
        
        % Changes  on the right MF of the new MF
        x(j).Rules( nearestMF , 1 ) = new_b;
        
        % update/sort the Rule Struct
        for i = n_rules:-1:nearestMF
            [x(j).Rules(i+1,:)] = x(j).Rules(i,:);
        end
        
        % Insert the new rule
        x(j).Rules(nearestMF, 1 ) = new_left;
        x(j).Rules(nearestMF, 2 ) = new_b;
        x(j).Rules(nearestMF, 3 ) = new_right;
        x(j).Rules(nearestMF, 4 ) = 0.0;
        
        rule = nearestMF;% index of the new rule
    end
  
end

end