% For use with Matlab
% Copyright (C) 2019 -- Jérôme Mendes <jermendes@gmail.com> or <jermendes@isr.uc.pt>, Rui Araújo <rui@isr.uc.pt>
% Jérôme Mendes, Francisco Souza, and Rui Araújo. "Online Evolving Fuzzy Control Design: An Application to a CSTR Plant". In Proc. IEEE 15th International Conference on Industrial Informatics (INDIN 2017), 2017.
% The "SEDFLC toolbox" comes with ABSOLUTELY NO WARRANTY;
% In case of publication of any application of this method, please, cite the work:
% Jérôme Mendes, Francisco Souza and Rui Araújo. 
% Online Evolving Fuzzy Control Design: An Application to a CSTR Plant.
% In Proc. IEEE 15th International Conference on Industrial Informatics (INDIN 2017), pages 1-8, Emden, Germany, July 24 - July 17
% DOI: https://doi.org/10.1109/INDIN.2017.8104774
% http://www.isr.uc.pt/~jermendes/    ;     http://www.isr.uc.pt/~rui/

function [command, x] = EvolveFLC(data_input, x, C, error, data_window, window_size, control_min_max)
%% Description
% This function is the core function to self-evolve the fuzzy controller

%% Function inputs 
% data_input -  values of the input variable 
% x - Struct with all fuzzy control parameters
% C - is the adaptation gain 
% error - is the tracking error, i.e. r(k)-y(k)
% data_window - temporal sliding window
% window_size - Size of the temporal sliding window 
% control_min_max - Universe of discourse of the control variable [min max]

%% Function output
% command -  fuzzy control signal - equation 8
% x - Struct with all fuzzy control parameters

%% Global variables
global summed_error_window dif_controller Evolve_Parameterts delta delta_dif_controller

%% Variables Definition
actual_size_window = size(data_window, 2); % Size of the temporal sliding window 
Add = 0;  % Auxiliary variable

%% Function to update the antecedent parameters (Algorithm 2 Step 3.a)
x = UpdateAntecedent(x, data_input);

%% Function to update the consequent parameters (Algorithm 2 Step 3.b)
x = UpdateConsequent(x, C, error, data_input, control_min_max); 

%% If the temporal sliding window is filled (Algorithm 2 Step 3.b)
if actual_size_window == window_size 
    
    %% Function to obtain the estimated control error (equation (19), Algorithm 2 Step 3.c.i)
    [error_controller_estimation_all, error_controller_estimation_local] = FuzzyController_EstimationError(x, data_window, window_size);

    error_controller_estimation = abs(sum(error_controller_estimation_all)); % estimated control error 
    
    dif_controller = [ dif_controller abs(error_controller_estimation - summed_error_window )]; % save the difference of the estimated control error (equation (24))
    
    %% If there are enough samples
    if size(dif_controller,2)>2
        
        delta_dif_controller = [delta_dif_controller dif_controller(1,end)-dif_controller(1,end-1)];

        %% If Criterion 1 is met (Algorithm 2 Step 3.c.ii)
        if  delta_dif_controller(1,end) > delta
            
            %% Find the input variable which has the maximal estimated control error
            j = find(error_controller_estimation_local == max(abs(error_controller_estimation_local)),1);

            %% This function obtains the position of the candidate membership function for variable j (Equation (22))
            [MF_center, nearestMF] = LocateMF_Gravity(error_controller_estimation_all, data_window, j,x);
            
            %% This function, if the Criterion 2 is met, adds a new  membership function on input variable j 
            [x, Add, rule] = AddMF_ErrorAVG(x, nearestMF,j, MF_center);
            
            %% If the candidate MF was added
            if Add==1
                %% This function obtains the consequent parameters of the new fuzzy rule (Equation (23))
                x = AddNewConsequent(x, j, rule,  data_window);
            end
            
        end
    end
    
    %% Save the estimated control error 
    summed_error_window = abs(sum(error_controller_estimation_all));
end

%% This function obtains the fuzzy control signal - equation 8
command = FuzzyController (data_input, x);

%% Save some importante parameters to do plots
Evolve_Parameterts.Add = [Evolve_Parameterts.Add Add]; % Flag Add 
Evolve_Parameterts.Sample = Evolve_Parameterts.Sample +1;
Evolve_Parameterts.dif_controller = dif_controller;% the difference of the estimated control error
Evolve_Parameterts.delta_dif_controller = delta_dif_controller;% the difference of difference of the estimated control error
for j=1:size(x,2)% for all input variables
    Evolve_Parameterts.nRules{j} = [Evolve_Parameterts.nRules{j} size(x(j).Rules,1)];% number of rules
    Evolve_Parameterts.MFs{j,Evolve_Parameterts.Sample} = x(j).Rules; % rules matrix (parameters)
end

end